import json
import os
from dataclasses import dataclass
from pathlib import Path

import pytest

TEST_URL = "https://kognic-test-url.com"
TEST_URL_2 = "https://kognic-test-url-2.com"
JSON_RESULT = {"id": "9ed7dasdasd-08ff-4ae1-8952-37e3a323eb08"}
BYTES_RESULT = json.dumps(JSON_RESULT).encode("utf-8")

DIR_PATH = Path(os.path.dirname(os.path.realpath(__file__)))
RANDOM_DATA_PATH = DIR_PATH / "resources/random-data.txt"
RANDOM_DATA_2_PATH = DIR_PATH / "resources/random-data-2.txt"


@dataclass
class GetSpec:
    url: str
    result_bytes: bytes
    result_json: dict


@dataclass
class PutSpec:
    url: str
    path: Path
    content: bytes


def read_file_content(file_path):
    with open(file_path, "rb") as f:
        return f.read()


RANDOM_DATA = read_file_content(RANDOM_DATA_PATH)
RANDOM_DATA_2 = read_file_content(RANDOM_DATA_2_PATH)


@pytest.fixture
def get_spec():
    return GetSpec(url=TEST_URL, result_bytes=BYTES_RESULT, result_json=JSON_RESULT)


@pytest.fixture
def put_spec():
    return PutSpec(url=TEST_URL, path=RANDOM_DATA_PATH, content=RANDOM_DATA)


@pytest.fixture
def put_spec_2():
    return PutSpec(url=TEST_URL_2, path=RANDOM_DATA_2_PATH, content=RANDOM_DATA_2)



from typing import List

from kognic.base_clients.models import BaseSerializer, PageMetadata, PaginatedResponse


class Task(BaseSerializer):
    task_name: str
    task_uuid: str


class User(BaseSerializer):
    user_name: str
    age: int
    tasks: List[Task]


user_dict_snake = {
    "user_name": "John Doe",
    "age": 42,
    "tasks": [{"task_name": "Task 1", "task_uuid": "uuid1"}, {"task_name": "Task 2", "task_uuid": "uuid2"}],
}

user_dict_camel = {
    "userName": "John Doe",
    "age": 42,
    "tasks": [{"taskName": "Task 1", "taskUuid": "uuid1"}, {"taskName": "Task 2", "taskUuid": "uuid2"}],
}

user_model = User(
    user_name="John Doe",
    age=42,
    tasks=[Task(task_name="Task 1", task_uuid="uuid1"), Task(task_name="Task 2", task_uuid="uuid2")],
)


def test_serialize_user_snake():
    user = User.from_json(user_dict_snake)
    assert user == user_model


def test_serialize_user_camel():
    user = User.from_json(user_dict_camel)
    assert user == user_model


# Should deserialize to camel case
def test_deserialize_user():
    user = user_model.to_dict()
    assert user == user_dict_camel


def test_parse_paginated_response_without_metadata():
    response_dict = {"data": [user_dict_camel]}
    response = PaginatedResponse.from_json(response_dict)
    assert response == PaginatedResponse(data=[user_dict_camel], metadata=PageMetadata())


def test_parse_paginated_response_without_cursor():
    response_dict = {"data": [user_dict_camel], "metadata": {}}
    response = PaginatedResponse.from_json(response_dict)
    assert response == PaginatedResponse(data=[user_dict_camel], metadata=PageMetadata())


def test_parse_paginated_response_with_cursor():
    response_dict = {"data": [user_dict_camel], "metadata": {"next_cursor_id": "abc-123"}}
    response = PaginatedResponse.from_json(response_dict)
    assert response == PaginatedResponse(data=[user_dict_camel], metadata=PageMetadata(next_cursor_id="abc-123"))



import time
from uuid import uuid4

import pytest
import responses
from authlib.oauth2.auth import OAuth2Token
from requests.exceptions import HTTPError

from kognic.base_clients.http_client import HttpClient

base_url = "https://api.mocked.com"
auth_host = "https://api.auth.mocked.com"

http_client = HttpClient(auth=None, host=base_url, auth_host=auth_host)


def mock_auth():
    responses.add(
        responses.POST,
        f"{auth_host}/v1/auth/oauth/token",
        status=200,
        json=OAuth2Token(
            {
                "access_token": "the-token",
                "expires_in": 200,
                "expires_at": time.time() + 200,
            }
        ),
    )


@responses.activate
def test_delete():
    mock_auth()

    resource_uuid = uuid4()
    responses.add(
        responses.DELETE,
        f"{base_url}/v1/resources/{resource_uuid}",
        status=200,
        json={"message": "success"},
    )
    response = http_client.delete(f"{base_url}/v1/resources/{resource_uuid}")
    assert response == {"message": "success"}


@responses.activate
def test_delete_discard():
    mock_auth()

    resource_uuid = uuid4()
    responses.add(
        responses.DELETE,
        f"{base_url}/v1/resources/{resource_uuid}",
        status=200,
        json={"message": "success"},
    )
    response = http_client.delete(f"{base_url}/v1/resources/{resource_uuid}", discard_response=True)
    assert response is None


@responses.activate
def test_delete_on_not_found():
    mock_auth()

    resource_uuid = uuid4()
    responses.add(
        responses.DELETE,
        f"{base_url}/v1/resources/{resource_uuid}",
        status=404,
        json={"message": "not found"},
    )
    with pytest.raises(HTTPError) as e:
        http_client.delete(f"{base_url}/v1/resources/{resource_uuid}")
    assert str(e.value).startswith("404 Client Error: Not Found for url")


@responses.activate
def test_delete_with_dryrun():
    mock_auth()

    resource_uuid = uuid4()
    responses.add(
        responses.DELETE,
        f"{base_url}/v1/resources/{resource_uuid}",
        status=200,
        json={"message": "success"},
        match=[responses.matchers.header_matcher({"X-Dryrun": ""})],
    )
    response = http_client.delete(f"{base_url}/v1/resources/{resource_uuid}", dryrun=True)
    assert response == {"message": "success"}



import asyncio
import logging
from pathlib import Path
from typing import BinaryIO, Dict, Mapping

import httpx

from kognic.base_clients.async_retry_support import request_with_retry
from kognic.base_clients.cloud_storage.upload_spec import Callback, UploadableData, UploadSpec

log = logging.getLogger(__name__)


class UploadHandler:
    def __init__(self, max_retry_attempts: int = 23, max_retry_wait_time: int = 60, timeout: int = 60) -> None:
        """
        :param max_retry_attempts: Max number of attempts to retry uploading a file to GCS.
        :param max_retry_wait_time:  Max with time before retrying an upload to GCS.
        :param timeout: Max time to wait for response from server.
        """
        self.max_num_retries = max_retry_attempts
        self.max_retry_wait_time = max_retry_wait_time  # seconds
        self.timeout = timeout  # seconds

    async def _upload_file(self, httpx_client: httpx.AsyncClient, upload_url: str, file: UploadableData, headers: Dict[str, str]) -> None:
        """
        Upload the file to GCS, retries if the upload fails with some specific status codes or timeouts.
        """
        await request_with_retry(
            httpx_client.put,
            self.max_num_retries,
            self.max_retry_wait_time,
            url=upload_url,
            content=file.read() if hasattr(file, "read") else file,
            headers=headers,
            timeout=self.timeout,
        )

    async def upload_files(self, upload_specs: Mapping[str, UploadSpec]) -> None:
        """
        Upload all files to cloud storage

        :param upload_specs: map between filename and details of what to upload where
        """

        async def create_task(client: httpx.AsyncClient, upload_spec: UploadSpec):
            if upload_spec.data is None and upload_spec.callback is None:
                await self._upload_from_local_file(client, upload_spec)
            elif upload_spec.data is not None:
                await self._upload_from_blob(client, upload_spec)
            elif upload_spec.callback is not None:
                await self._upload_from_callback(client, upload_spec)

        async with httpx.AsyncClient() as httpx_client:
            tasks = [create_task(httpx_client, upload_spec) for upload_spec in upload_specs.values()]
            await asyncio.gather(*tasks)

    async def upload_file(self, file: BinaryIO, url: str) -> None:
        headers = {"Content-Type": "application/json"}
        async with httpx.AsyncClient() as httpx_client:
            await self._upload_file(httpx_client, url, file, headers)

    async def _upload(self, httpx_client: httpx.AsyncClient, upload_spec: UploadSpec, data: UploadableData):
        headers = {"Content-Type": upload_spec.content_type}
        await self._upload_file(httpx_client, upload_spec.destination, data, headers)

    async def _upload_from_blob(self, httpx_client: httpx.AsyncClient, upload_spec: UploadSpec):
        log.debug(f"Blob upload for filename={upload_spec.filename}")
        await self._upload(httpx_client, upload_spec, upload_spec.data)

    async def _upload_from_local_file(self, httpx_client: httpx.AsyncClient, upload_spec: UploadSpec):
        log.debug(f"Upload from local file for filename={upload_spec.filename}")
        with open(Path(upload_spec.filename).expanduser(), "rb") as file:
            file_content = file.read()
        await self._upload(httpx_client, upload_spec, file_content)

    async def _upload_from_callback(self, httpx_client: httpx.AsyncClient, upload_spec: UploadSpec):
        log.debug(f"Upload from callback for filename={upload_spec.filename}")
        try:
            if self._is_async(upload_spec.callback):
                data = await upload_spec.callback(upload_spec.filename)
            else:
                data = upload_spec.callback(upload_spec.filename)
            await self._upload(httpx_client, upload_spec, data)
        except Exception as e:
            raise RuntimeError("Failed to upload file: callback failed", e)

    @staticmethod
    def _is_async(callback: Callback) -> bool:
        return asyncio.iscoroutinefunction(callback)



import logging
import time
from typing import Callable, Optional

from requests import ConnectionError, HTTPError, RequestException, Response, Timeout

from kognic.base_clients.util import RETRYABLE_STATUS_CODES, get_wait_time

log = logging.getLogger(__name__)

HTTPErrorHandler = Callable[[HTTPError], None]


def request_with_retry(
    method: Callable[..., Response],
    max_num_retries: int,
    max_retry_wait_time: int,
    http_error_handler: Optional[HTTPErrorHandler] = None,
    *args,
    **kwargs,
) -> Response:
    return _request_with_retry(method, max_num_retries, max_num_retries, max_retry_wait_time, http_error_handler, *args, **kwargs)


def _request_with_retry(
    method: Callable[..., Response],
    number_of_retries: int,
    max_num_retries: int,
    max_retry_wait_time: int,
    http_error_handler: Optional[HTTPErrorHandler] = None,
    *args,
    **kwargs,
) -> Response:
    """
    Retrying HTTP request with exponential backoff.

    :param method: HTTP method executor (probably something like requests.put, auth_session.put).
    :param number_of_retries: Remaining number of retries
    :param max_num_retries: Max number of retry attempts
    :param max_retry_wait_time: Upper bound on wait time between retries (seconds)
    :param args: Named args for `method`
    :param kwargs: kwargs for `method`
    :return: Response
    """

    def handle_error(error_type: str, re: RequestException, retries_remaining: int):
        if retries_remaining > 0:
            _generic_error(error_type, retries_remaining, max_num_retries, max_retry_wait_time)
            return _request_with_retry(
                method, retries_remaining - 1, max_num_retries, max_retry_wait_time, http_error_handler, *args, **kwargs
            )
        else:
            log.error("Request error, no more retries left", re)
            raise re

    try:
        resp = method(*args, **kwargs)
        resp.raise_for_status()
        return resp
    except HTTPError as e:
        if e.response.status_code in RETRYABLE_STATUS_CODES:
            return handle_error("Retryable HTTP error", e, number_of_retries)
        elif http_error_handler is not None:
            http_error_handler(e)
        else:
            raise e
    except Timeout as e:
        return handle_error("Network timeout", e, number_of_retries)
    except ConnectionError as e:
        return handle_error("Connection error", e, number_of_retries)


def _generic_error(error_type: str, number_of_retries: int, max_num_retries: int, max_retry_wait_time: int) -> None:
    upload_attempt = max_num_retries - number_of_retries + 1
    wait_time = get_wait_time(upload_attempt, max_retry_wait_time)
    log.warning(f"Failed to upload file. {error_type}" f"Attempt {upload_attempt}/{max_num_retries}, retrying in {int(wait_time)} seconds.")
    time.sleep(wait_time)



from io import IOBase
from typing import Awaitable, BinaryIO, Callable, Optional, Union

UploadableData = Union[bytes, BinaryIO, IOBase]
SyncCallback = Callable[[str], UploadableData]
AsyncCallback = Callable[[str], Awaitable[UploadableData]]
Callback = Union[SyncCallback, AsyncCallback]


class UploadSpec:
    """
    Represents an upload to be performed to some resource ID: what is to be uploaded and where.
    A filename must always be set, and a data blob or callback may optionally be set. With just a filename this upload
    will use the named file. With a filename and a blob, the data blob is used and the filename is informational only.
    If a callback is set, it will be called with the filename as its argument and must return a blob (bytes) for upload.
    """

    destination: str
    filename: str
    content_type: str
    data: Optional[UploadableData]
    callback: Optional[Callback]

    def __init__(
        self,
        destination: str,
        content_type: str,
        filename: str = None,
        data: Optional[UploadableData] = None,
        callback: Optional[Callback] = None,
    ):
        self.destination = destination
        self.filename = filename
        self.content_type = content_type
        self.data = data
        self.callback = callback

    def __repr__(self):
        data_part = "" if self.data is None else f", data"
        cb_part = "" if self.callback is None else f", callback"
        return f"UploadSpec({self.content_type}, destination={self.destination}, filename={self.filename}{data_part}{cb_part})"



from pathlib import Path

import aiofiles
import pytest
from pytest_httpx import HTTPXMock

from kognic.base_clients.cloud_storage import UploadSpec
from kognic.base_clients.cloud_storage.upload_handler import UploadHandler

pytest_plugins = ("pytest_asyncio",)


@pytest.mark.asyncio
class TestAsyncUploadHandler:
    async def test_upload_file_200(self, httpx_mock: HTTPXMock, put_spec):
        httpx_mock.add_response(method="PUT", status_code=200, url=put_spec.url, match_content=put_spec.content)
        async_upload_handler = UploadHandler(max_retry_attempts=3, max_retry_wait_time=0.1)
        with open(put_spec.path, "rb") as f:
            await async_upload_handler.upload_file(f, put_spec.url)
        assert len(httpx_mock.get_requests()) == 1

    async def test_upload_file_500_then_200(self, httpx_mock: HTTPXMock, put_spec):
        httpx_mock.add_response(method="PUT", status_code=500, url=put_spec.url, match_content=put_spec.content)
        httpx_mock.add_response(method="PUT", status_code=500, url=put_spec.url, match_content=put_spec.content)
        httpx_mock.add_response(method="PUT", status_code=200, url=put_spec.url, match_content=put_spec.content)
        async_upload_handler = UploadHandler(max_retry_attempts=3, max_retry_wait_time=0.1)
        with open(put_spec.path, "rb") as f:
            await async_upload_handler.upload_file(f, put_spec.url)
        assert len(httpx_mock.get_requests()) == 3

    async def test_upload_files_local_200(self, httpx_mock: HTTPXMock, put_spec, put_spec_2):
        upload_specs = {
            "random-data.txt": UploadSpec(
                destination=put_spec.url,
                filename=str(put_spec.path),
                content_type="application/octet-stream",
            ),
            "random-data-2.txt": UploadSpec(
                destination=put_spec_2.url,
                filename=str(put_spec_2.path),
                content_type="application/octet-stream",
            ),
        }

        httpx_mock.add_response(method="PUT", status_code=200, url=put_spec.url, match_content=put_spec.content)
        httpx_mock.add_response(method="PUT", status_code=200, url=put_spec_2.url, match_content=put_spec_2.content)
        async_upload_handler = UploadHandler()

        await async_upload_handler.upload_files(upload_specs=upload_specs)
        assert len(httpx_mock.get_requests()) == 2

    async def test_upload_files_file_data_200(self, httpx_mock: HTTPXMock, put_spec, put_spec_2):
        file_content = put_spec.content
        file_content_2 = put_spec_2.content
        upload_specs = {
            "random-data.txt": UploadSpec(
                destination=put_spec.url,
                filename=str(put_spec.path),
                content_type="application/octet-stream",
                data=file_content,
            ),
            "random-data-2.txt": UploadSpec(
                destination=put_spec_2.url,
                filename=str(put_spec_2.path),
                content_type="application/octet-stream",
                data=file_content_2,
            ),
        }

        httpx_mock.add_response(method="PUT", status_code=200, url=put_spec.url, match_content=file_content)
        httpx_mock.add_response(method="PUT", status_code=200, url=put_spec_2.url, match_content=file_content_2)

        async_upload_handler = UploadHandler()

        await async_upload_handler.upload_files(upload_specs=upload_specs)
        assert len(httpx_mock.get_requests()) == 2

    async def test_upload_files_callback_200(self, httpx_mock: HTTPXMock, put_spec, put_spec_2):
        upload_specs = {
            "random-data.txt": UploadSpec(
                destination=put_spec.url,
                filename=str(put_spec.path),
                content_type="application/octet-stream",
                callback=lambda fname: Path(fname).open("rb"),
            ),
            "random-data-2.txt": UploadSpec(
                destination=put_spec_2.url,
                filename=str(put_spec_2.path),
                content_type="application/octet-stream",
                callback=lambda fname: Path(fname).open("rb").read(),
            ),
        }

        httpx_mock.add_response(method="PUT", status_code=200, url=put_spec.url, match_content=put_spec.content)
        httpx_mock.add_response(method="PUT", status_code=200, url=put_spec_2.url, match_content=put_spec_2.content)

        async_upload_handler = UploadHandler()

        await async_upload_handler.upload_files(upload_specs=upload_specs)
        assert len(httpx_mock.get_requests()) == 2

    async def test_upload_files_async_callback_200(self, httpx_mock: HTTPXMock, put_spec, put_spec_2):
        async def callback(fname: str) -> bytes:
            async with aiofiles.open(fname, "rb") as f:
                return await f.read()

        upload_specs = {
            "random-data.txt": UploadSpec(
                destination=put_spec.url,
                filename=str(put_spec.path),
                content_type="application/octet-stream",
                callback=callback,
            ),
            "random-data-2.txt": UploadSpec(
                destination=put_spec_2.url,
                filename=str(put_spec_2.path),
                content_type="application/octet-stream",
                callback=callback,
            ),
        }

        httpx_mock.add_response(method="PUT", status_code=200, url=put_spec.url, match_content=put_spec.content)
        httpx_mock.add_response(method="PUT", status_code=200, url=put_spec_2.url, match_content=put_spec_2.content)

        async_upload_handler = UploadHandler()

        await async_upload_handler.upload_files(upload_specs=upload_specs)
        assert len(httpx_mock.get_requests()) == 2

    async def test_upload_files_when_async_callback_failed(self, httpx_mock: HTTPXMock, put_spec, put_spec_2):
        async def callback(fname: str) -> bytes:
            async with aiofiles.open("I-don't-exist.txt", "rb") as f:
                return await f.read()

        upload_specs = {
            "random-data.txt": UploadSpec(
                destination=put_spec.url,
                filename=str(put_spec.path),
                content_type="application/octet-stream",
                callback=lambda fname: Path(fname).open("rb"),
            ),
            "random-data-2.txt": UploadSpec(
                destination=put_spec_2.url,
                filename=str(put_spec_2.path),
                content_type="application/octet-stream",
                callback=callback,
            ),
        }

        httpx_mock.add_response(method="PUT", status_code=200, url=put_spec.url, match_content=put_spec.content)

        async_upload_handler = UploadHandler()

        with pytest.raises(RuntimeError) as exc_info:
            await async_upload_handler.upload_files(upload_specs=upload_specs)
            assert exc_info.value.args[0] == "Failed to upload file: callback failed"
            assert isinstance(exc_info.value.args[1], FileNotFoundError)

        assert len(httpx_mock.get_requests()) == 1
